# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
import keras



# Copied from Keras-CV
# https://github.com/keras-team/keras-cv/blob/master/keras_cv/api_export.py
def maybe_register_serializable(symbol, package):
    if isinstance(symbol, types.FunctionType) or hasattr(symbol, "get_config"):
        keras.saving.register_keras_serializable(package=package)(symbol)


# if namex:
#     class teras_export(namex.export):
#         def __init__(self, path, package="teras"):
#             super().__init__(package="teras", path=path)
#             self.package = package
#
#         def __call__(self, symbol):
#             maybe_register_serializable(symbol, self.package)
#             return super().__call__(symbol)
#
# else:
class teras_export:
    def __init__(self, path, package="teras"):
        self.package = package

    def __call__(self, symbol):
        maybe_register_serializable(symbol, self.package)
        return symbol



from teras._src.preprocessing.data_transformers.ctgan import CTGANDataTransformer as _BaseDataTransformer
from teras._src.typing import FeaturesNamesType


class TVAEDataTransformer(_BaseDataTransformer):
    """
    TVAEDataTransformer class that is exactly similar to the
    CTGANDataTransformer, it just acts as a wrapper for convenience.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        categorical_features: list, List of categorical features names in the
            dataset.
        continuous_features: list, List of continuous features names in the
            dataset.
        max_clusters: int, Maximum Number of clusters to use in
            `ModeSpecificNormalization`. Defaults to 10.
        std_multiplier: int, Multiplies the standard deviation in the
            normalization. Defaults to 4.
        weight_threshold: float, The minimum value a component weight can
            take to be considered a valid component. `weights_` under this
            value will be ignored. (Taken from the official implementation.)
            Defaults to 0.005.
        covariance_type: str, Parameter for the `GaussianMixtureModel`
            class of sklearn. Defaults to "full".
        weight_concentration_prior_type: str, Parameter for the
            `GaussianMixtureModel` class of sklearn. Defaults to
            "dirichlet_process"
        weight_concentration_prior: float, Parameter for the
            `GaussianMixtureModel` class of sklearn. Defaults to 0.001.
    """
    def __init__(self,
                 continuous_features: FeaturesNamesType = None,
                 categorical_features: FeaturesNamesType = None,
                 max_clusters: int = 10,
                 std_multiplier: int = 4,
                 weight_threshold: float = 0.005,
                 covariance_type: str = "full",
                 weight_concentration_prior_type: str = "dirichlet_process",
                 weight_concentration_prior: float = 0.001
                 ):
        super().__init__(
            continuous_features=continuous_features,
            categorical_features=categorical_features,
            max_clusters=max_clusters,
            std_multiplier=std_multiplier,
            weight_threshold=weight_threshold,
            covariance_type=covariance_type,
            weight_concentration_prior_type=weight_concentration_prior_type,
            weight_concentration_prior=weight_concentration_prior)



import keras

from teras._src.layers.layer_list import LayerList
from teras._src.typing import IntegerSequence
from teras._src.api_export import teras_export


@teras_export("teras.models.TVAEDecoder")
class TVAEDecoder(keras.Model):
    """
    Encoder for the TVAE model as proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        data_dim: int, Dimensionality of the input dataset.
        decompression_dims: A sequence of integers. For each value in
            the sequence, a dense layer of that dimensionality is added to
            construct a decompression block.
    """
    def __init__(self,
                 data_dim: int,
                 decompression_dims: IntegerSequence = (128, 128),
                 **kwargs):
        super().__init__(**kwargs)

        if not isinstance(decompression_dims, (list, tuple)):
            raise ValueError(
                f"`decompression_dims` must be a sequence of integers. "
                f"Received: {decompression_dims}")

        self.data_dim = data_dim
        self.decompression_dims = decompression_dims

        self.compression_block = []
        for i, units in enumerate(self.decompression_dims, start=1):
            self.compression_block.append(
                keras.layers.Dense(units=units,
                                   activation="relu",
                                   name=f"decompression_layer_{i}"))
        self.decompression_block = LayerList(
            self.compression_block,
            sequential=True,
            name="tvae_encoder_compression_block"
        )
        self.projection_layer = keras.layers.Dense(self.data_dim,
                                                   name="projection_layer")
        self.sigmas = self.add_weight(shape=(self.data_dim,),
                                      initializer="ones", trainable=True,
                                      name="sigmas") * 0.1

    def build(self, input_shape):
        self.decompression_block.build(input_shape)
        input_shape = self.decompression_block.compute_output_shape(input_shape)
        self.projection_layer.build(input_shape)

    def call(self, inputs):
        x_generated = self.projection_layer(self.decompression_block(inputs))
        return x_generated, self.sigmas

    def predict_step(self, z):
        generated_samples, _ = self(z)
        return generated_samples

    def compute_output_shape(self, input_shape):
        batch_size, input_dim = input_shape
        return ((batch_size, self.data_dim),
                (self.data_dim,))

    def get_config(self):
        config = super().get_config()
        config.update({
            'data_dim': self.data_dim,
            'decompression_dims': self.decompression_dims
        })
        return config




from .utils import dataframe_to_tf_dataset
import pandas as pd
import numpy as np


def test_dataframe_to_tf_dataset_without_label():
    df = pd.DataFrame(data={"length": np.ones(10),
                            "area": np.ones(10) * 2})
    ds = dataframe_to_tf_dataset(df)


def test_dataframe_to_tf_dataset_with_single_label():
    df = pd.DataFrame(data={"length": np.ones(10),
                            "area": np.ones(10) * 2})
    ds = dataframe_to_tf_dataset(df, target="area")


def test_dataframe_to_tf_dataset_with_multiple_labels():
    df = pd.DataFrame(data={"length": np.ones(10),
                            "area": np.ones(10) * 2,
                            "perimeter": np.ones(10) * 3})
    ds = dataframe_to_tf_dataset(df, target=["area", "perimeter"])



import keras
from teras._src.layers.transformer.encoder_layer import TransformerEncoderLayer
from teras._src.layers.transformer.feedforward import TransformerFeedForward
from teras._src.layers.saint.multi_head_inter_sample_attention import SAINTMultiHeadInterSampleAttention
from teras._src.api_export import teras_export


@teras_export("teras.layers.SAINTEncoderLayer")
class SAINTEncoderLayer(keras.layers.Layer):
    """
    SAINTEncoderLayer layer as proposed in the paper,
    "SAINT: Improved Neural Networks for Tabular Data".

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        embedding_dim: int, dimensionality of the embeddings
        num_heads: int, number of attention heads to use in the
            `MultiHeadAttention` layer.
        feedforward_dim: int, hidden dimensionality to use in the
            `TransformerFeedForward` layer.
        attention_dropout: float, dropout value to use in the
            `MultiHeadAttention` layer. Defaults to 0.
        feedforward_dropout: float, dropout value to use in the
            `TransformerFeedForward` layer. Defaults to 0.
        layer_norm_epsilon: float, epsilon value to use in the
            `LayerNormalization` layer. Defaults to 1e-5.
    """
    def __init__(self,
                 embedding_dim: int,
                 num_heads: int = 8,
                 feedforward_dim: int = None,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 layer_norm_epsilon: float = 1e-5,
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.feedforward_dim = feedforward_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.layer_norm_epsilon = layer_norm_epsilon

        # ====== Self Attention Block ========
        self.self_attention_block = TransformerEncoderLayer(
            embedding_dim=self.embedding_dim,
            num_heads=self.num_heads,
            feedforward_dim=self.feedforward_dim,
            attention_dropout=self.attention_dropout,
            feedforward_dropout=self.feedforward_dropout,
            layer_norm_epsilon=self.layer_norm_epsilon,
            name="Self_Attention_Block",
        )

        # ====== Inter Sample Attention Block ========
        self.inter_sample_attention = SAINTMultiHeadInterSampleAttention(
            num_heads=self.num_heads,
            key_dim=self.embedding_dim // self.num_heads,   # ref: paper
            value_dim=None,
            dropout=self.attention_dropout,
        )
        self.isab_feed_forward = TransformerFeedForward(
            embedding_dim=self.embedding_dim,
            hidden_dim=self.feedforward_dim,
            activation="gelu",
            dropout=feedforward_dropout
        )
        self.isab_norm_1 = keras.layers.LayerNormalization(
            epsilon=self.layer_norm_epsilon,
        )
        self.isab_norm_2 = keras.layers.LayerNormalization(
            epsilon=self.layer_norm_epsilon,
        )

    def build(self, input_shape):
        self.self_attention_block.build(input_shape)
        self.inter_sample_attention.build(input_shape)
        self.isab_feed_forward.build(input_shape)
        self.isab_norm_1.build(input_shape)
        self.isab_norm_2.build(input_shape)

    def call(self, inputs):
        # ====== Self Attention Block ========
        x = self.self_attention_block(inputs)
        # ====== Inter Sample Attention Block ========
        residue = x
        x = self.inter_sample_attention(x)
        x = keras.layers.add([x, residue])
        x = self.isab_norm_1(x)
        residue = x
        x = self.isab_feed_forward(x)
        x = keras.layers.add([x, residue])
        x = self.isab_norm_2(x)
        return x

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            "embedding_dim": self.embedding_dim,
            "num_heads": self.num_heads,
            "feedforward_dim": self.feedforward_dim,
            "attention_dropout": self.attention_dropout,
            "feedforward_dropout": self.feedforward_dropout,
            "layer_norm_epsilon": self.layer_norm_epsilon,
        })



import keras
from keras import ops
from teras._src.api_export import teras_export
from teras._src.layers.layer_list import LayerList


@teras_export("teras.layers.SAINTEmbedding")
class SAINTEmbedding(keras.layers.Layer):
    """
    SAINTEmbedding layer as proposed in the paper,
    "SAINT: Improved Neural Networks for Tabular Data".

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        embedding_dim: int, dimensionality of the embeddings
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use the value `0` as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.

    Shapes:
        Input Shape: (batch_size, input_dim)
        Output Shape: (batch_size, input_dim, embedding_dim)
    """
    def __init__(self,
                 embedding_dim: int,
                 cardinalities: list,
                 **kwargs):
        super().__init__(**kwargs)
        self.cardinalities = cardinalities
        self.embedding_dim = embedding_dim

        self.embedding_layers = []
        for card in self.cardinalities:
            if card == 0:
                # it's continuous
                self.embedding_layers.append(
                    keras.layers.Dense(self.embedding_dim,
                                       activation="relu")
                )
            else:
                # it's categorical
                self.embedding_layers.append(
                    keras.layers.Embedding(
                        input_dim=card + 1,
                        output_dim=self.embedding_dim))
        self.embedding_layers = LayerList(self.embedding_layers,
                                          sequential=False)

    def build(self, input_shape):
        # since each embedding layer only operates on a single feature
        self.embedding_layers.build((input_shape[0], 1))

    def compute_output_shape(self, input_shape):
        return input_shape + (self.embedding_dim,)

    def call(self, inputs):
        feature = ops.take(inputs, indices=[0], axis=1)
        # As much as I'd like to use the empty tensor, i can't because
        # every framework uses different methods for assigning values in
        # an **efficient** way. What's efficient in one framework isn't
        # efficient in the other. I ain't making a spaghetti. I'd rather do
        # it this way.
        embeddings = self.embedding_layers[0](feature)
        if len(ops.shape(embeddings)) == 2:
            embeddings = ops.expand_dims(embeddings, axis=1)
        for idx, embedding_layer in enumerate(self.embedding_layers[1:],
                                              start=1):
            feature = ops.take(inputs, indices=[idx], axis=1)
            feature_embeddings = embedding_layer(feature)
            if len(ops.shape(feature_embeddings)) == 2:
                feature_embeddings = ops.expand_dims(
                    feature_embeddings, axis=1)
            embeddings = ops.concatenate([embeddings, feature_embeddings],
                                         axis=1)
        return embeddings



import keras
import numpy as np
from keras import random, ops, backend
from teras._src.layers.layer_list import LayerList


class BaseTabTransformerMLMPretrainer(keras.Model):
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 missing_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.data_dim = data_dim
        if missing_rate < 0. or missing_rate >= 1.0:
            raise ValueError(
                f"`k` must be in range [0, 1). Received {missing_rate}"
            )
        self.missing_rate = missing_rate
        self.mask_seed = mask_seed
        self._numpy_rng = np.random.default_rng(self.mask_seed)
        self._seed_for_mask = random.SeedGenerator(self.mask_seed)

        self.features_predictor = keras.layers.Dense(
            units=data_dim,
            name="features_predictor")

        self._pretrained = False

    def build(self, input_shape):
        self.model.build(input_shape)
        input_shape = self.model.compute_output_shape(input_shape)
        self.features_predictor.build(input_shape)

    def compile(self,
                loss=keras.losses.CategoricalCrossentropy(
                    from_logits=True),
                optimizer=keras.optimizers.Adam(5e-3),
                **kwargs):
        super().compile(loss=loss, optimizer=optimizer, **kwargs)

    def call(self, inputs, mask, **kwargs):
        x = inputs * mask
        x = self.model(x)
        x = self.features_predictor(x)
        x = x * (1 - mask)
        return x

    def _create_mask(self, input_shape):
        if len(input_shape) != 2:
            raise ValueError(
                "Expected `input_shape` to have a rank of 2. "
                f"Received, {input_shape} with rank {len(input_shape)}")
        batch_size, num_features = input_shape
        num_features_to_miss = int(self.missing_rate * num_features)
        # mask = np.zeros(input_shape)
        mask = ops.zeros(input_shape)
        mask = ops.convert_to_numpy(mask)
        features_idx = np.arange(num_features)

        for i in range(batch_size):
            features_idx_to_replace = self._numpy_rng.choice(
                features_idx,
                num_features_to_miss,
                replace=False)
            mask[i, features_idx_to_replace] = 1.
        mask = ops.convert_to_tensor(mask, dtype="int32")
        return mask

    def get_config(self):
        config = {
            "name": self.name,
            "trainable": self.trainable,
            "model": keras.layers.serialize(self.model),
            "data_dim": self.data_dim,
            "missing_rate": self.missing_rate,
            "mask_seed": self.mask_seed
        }
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        return cls(model=model, **config)

    @property
    def pretrained_model(self):
        if not self._pretrained:
            raise AssertionError(
                "The `fit` method of the `TabTransformerMLMPretrainer` "
                "has not yet been called. Please train the it before "
                "accessing the `pretrained_model` attribute."
            )
        return self.model


class BaseTabTransformerRTDPretrainer(keras.Model):
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 replace_rate: float = 0.3,
                 mask_seed: int = 1337,
                 shuffle_seed: int = 1999,
                 **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.data_dim = data_dim
        # here `k` represents replace rate instead of missing rate like
        # in the MLM version of the pretrainer
        if replace_rate < 0. or replace_rate >= 1.0:
            raise ValueError(
                f"`k` must be in range [0, 1). Received {replace_rate}"
            )
        self.replace_rate = replace_rate
        self.mask_seed = mask_seed
        self._numpy_rng = np.random.default_rng(self.mask_seed)
        self._seed_for_mask = random.SeedGenerator(self.mask_seed)
        self.shuffle_seed = shuffle_seed
        self._seed_for_shuffling = random.SeedGenerator(self.shuffle_seed)

        # binary predictors, one for each feature, which predict if the
        # value of the features is replaced or not
        self.predictors = LayerList([
            keras.layers.Dense(1,  name=f"feature_{i}_predictor")
            for i in range(data_dim)
        ],
            sequential=False)

    def build(self, input_shape):
        self.model.build(input_shape)
        input_shape = self.model.compute_output_shape(input_shape)
        self.predictors.build(input_shape)

    def compile(self,
                loss=keras.losses.BinaryCrossentropy(
                    from_logits=True),
                optimizer=keras.optimizers.Adam(5e-3),
                **kwargs):
        super().compile(loss=loss, optimizer=optimizer, **kwargs)

    def call(self, inputs, mask, **kwargs):
        # Since in RTD, for a sample, we randomly replace k% of its
        # features values using random values of those features.
        # We can efficiently achieve this by first getting
        # x_rand = shuffle(inputs)
        # then, to apply replacement,
        # inputs = (inputs * (1-mask)) + (x_rand * mask)
        inputs_shuffled = random.shuffle(inputs,
                                         axis=0,
                                         seed=self._seed_for_shuffling)
        inputs = inputs * mask + (inputs_shuffled * (1 - mask))
        x = self.model(inputs)
        predictions = self.predictors[0](x)
        for i in range(1, self.data_dim):
            pred = self.predictors[i](x)
            predictions = ops.concatenate([predictions, pred], axis=1)
        return predictions

    def _create_mask(self, input_shape):
        if len(input_shape) != 2:
            raise ValueError(
                "Expected `input_shape` to have a rank of 2. "
                f"Received, {input_shape} with rank {len(input_shape)}")
        batch_size, num_features = input_shape
        num_features_to_replace = int(self.replace_rate * num_features)
        mask = np.zeros(input_shape)
        # mask = ops.zeros(input_shape)
        # mask = np.asarray(mask)
        features_idx = np.arange(num_features)

        for i in range(batch_size):
            features_idx_to_replace = self._numpy_rng.choice(
                features_idx,
                num_features_to_replace,
                replace=False)
            mask[i, features_idx_to_replace] = 1.
        mask = ops.convert_to_tensor(mask, dtype="int32")
        return mask

    def get_config(self):
        config = {
            "name": self.name,
            "trainable": self.trainable,
            "model": keras.layers.serialize(self.model),
            "data_dim": self.data_dim,
            "replace_rate": self.replace_rate,
            "mask_seed": self.mask_seed,
            "shuffle_seed": self.shuffle_seed
        }
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        return cls(model=model, **config)

    @property
    def pretrained_model(self):
        if not self._pretrained:
            raise AssertionError(
                "The `fit` method of the `TabTransformerRTDPretrainer` "
                "has not yet been called. Please train the it before "
                "accessing the `pretrained_model` attribute."
            )
        return self.model



from keras import random, ops
from teras._src.layers.tabnet.feature_transformer import TabNetFeatureTransformer
from keras.src.testing.test_case import TestCase


class TabNetFeatureTransformerTest(TestCase):
    def setUp(self):
        self.batch_size = 16
        self.input_batch = random.normal((self.batch_size, 5))

    def test_valid_output_shape(self):
        layer = TabNetFeatureTransformer(hidden_dim=8,
                                         num_shared_layers=2,
                                         num_decision_dependent_layers=2,
                                         batch_momentum=0.99)
        outputs = layer(self.input_batch)
        self.assertEqual(ops.shape(outputs), (self.batch_size, 8))




import keras
from teras._src.layers.transformer.feedforward import TransformerFeedForward
from teras._src.api_export import teras_export


@teras_export("teras.layers.TransformerEncoderLayer")
class TransformerEncoderLayer(keras.layers.Layer):
    """
    Transformer Encoder Layer as proposed in the original Transformer
    architecture in the "Attention is all you need" paper.

    This is the layer that makes up the encoder in the architecture.
    This is made up of `MultiHeadAttention` and `TransformerFeedForward`
    layers.

    Reference(s):
        https://arxiv.org/abs/1706.03762

    Args:
        embedding_dim: int, dimensionality of the embeddings used
            by the model. It is also referred to as the `d_model` or
            model dimensionality.
        num_heads: int, number of attention heads to use in the
            `MultiHeadAttention` layer.
        feedforward_dim: int, hidden dimensionality to use in the
            `TransformerFeedForward` layer.
        attention_dropout: float, dropout value to use in the
        `MultiHeadAttention` layer. Defaults to 0.
        feedforward_dropout: float, dropout value to use in the
            `TransformerFeedForward` layer. Defaults to 0.
        layer_norm_epsilon: float, epsilon value to use in the
            `LayerNormalization` layer. Defaults to 1e-5.
        use_normalization: bool, whether to use `LayerNormalization`.
            In some architecture, normalization isn't applied to the
            very first layer, so to accomodate such architectures,
            we introduced this parameter.
            Defaults to `True`.
        pre_normalization: bool, whether to use Pre-Normalization technique
            whereby `LayerNormalization` is applied to inputs of the
            `MultiHeadAttention` or `FeedForward` and then outputs of
            those layers are elementwise added to the original inputs.
            Defaults to `False`, as the original Transformers architecture
            doesn't use pre-normalization.

    Shapes:
        Input Shape: `(batch_size, num_features, embedding_dim)`
        Output Shape: `(batch_size, num_features, embedding_dim)`
    """
    def __init__(self,
                 embedding_dim: int,
                 num_heads: int = 8,
                 feedforward_dim: int = None,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 layer_norm_epsilon: float = 1e-5,
                 use_normalization: bool = True,
                 pre_normalization: bool = False,
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.feedforward_dim = feedforward_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.layer_norm_epsilon = layer_norm_epsilon
        self.use_normalization = use_normalization
        self.pre_normalization = pre_normalization

        self.attention = keras.layers.MultiHeadAttention(
            num_heads=self.num_heads,
            key_dim=self.embedding_dim,
            dropout=attention_dropout
        )
        self.feedforward = TransformerFeedForward(
            embedding_dim=self.embedding_dim,
            hidden_dim=self.feedforward_dim,
            dropout=self.feedforward_dropout
        )
        self.add_1 = keras.layers.Add()
        self.add_2 = keras.layers.Add()
        if self.use_normalization:
            self.layer_norm_1 = keras.layers.LayerNormalization(
                epsilon=self.layer_norm_epsilon
            )
            self.layer_norm_2 = keras.layers.LayerNormalization(
                epsilon=self.layer_norm_epsilon
            )

    def build(self, input_shape):
        self.feedforward.build(input_shape)
        if self.use_normalization:
            self.layer_norm_1.build(input_shape)
            self.layer_norm_2.build(input_shape)

    def call(self, inputs):
        residue = inputs
        if self.use_normalization and self.pre_normalization:
            x = self.layer_norm_1(inputs)
            x = self.attention(x, x)
            x = self.add_1([x, residue])
            residue = x
            x = self.layer_norm_2(x)
            x = self.feedforward(x)
            x = self.add_2([x, residue])
        else:
            x = self.attention(inputs, inputs)
            x = self.add_1([x, residue])
            if self.use_normalization:
                x = self.layer_norm_1(x)
            residue = x
            x = self.feedforward(x)
            x = self.add_2([x, residue])
            if self.use_normalization:
                x = self.layer_norm_2(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update({
            "embedding_dim": self.embedding_dim,
            "num_heads": self.num_heads,
            "feedforward_dim": self.feedforward_dim,
            "attention_dropout": self.attention_dropout,
            "feedforward_dropout": self.feedforward_dropout,
            "layer_norm_epsilon": self.layer_norm_epsilon,
            "use_normalization": self.use_normalization,
            "pre_normalization": self.pre_normalization
        })
        return config



import keras
from keras import random, ops
from teras._src.api_export import teras_export
from keras import backend


@teras_export("teras.layers.CutMix")
class CutMix(keras.layers.Layer):
    """
    CutMix is a regularization layer proposed by Sangdoo Yun et al.
    in the paper,
    CutMix: Regularization Strategy to Train Strong Classifiers
    with Localizable Features

    It was originally proposed for image data but here it has been
    adapted for Tabular data.

    Args:
        probability: float, CutMix probability which is used in
            generation of mask that is used to mix samples together.
            Defaults to 0.3
        mask_seed: int, seed used in the generation fo the mask
            Defaults to 1337
        shuffle_seed: int, seed used in shuffling the inputs
            Defaults to 1999
    """
    def __init__(self,
                 probability: float = 0.3,
                 mask_seed: int = 1337,
                 shuffle_seed: int = 1999,
                 **kwargs):
        super().__init__(**kwargs)
        self.probability = probability
        self.mask_seed = mask_seed
        self._mask_seed_gen = random.SeedGenerator(self.mask_seed)
        self.shuffle_seed = shuffle_seed
        self._shuffle_seed_gen = random.SeedGenerator(self.shuffle_seed)

    def build(self, input_shape):
        # there's nothing to build lol
        pass

    def call(self, inputs):
        # Generate mask for CutMix mixing
        mask_cutmix = random.binomial(ops.shape(inputs),
                                      counts=1,
                                      probabilities=self.probability,
                                      seed=self._mask_seed_gen)

        # For each data sample select a partner to mix it with at random.
        # To efficiently achieve this, we can just shuffle the data
        random_partners = random.shuffle(inputs,
                                         axis=0,
                                         seed=self._shuffle_seed_gen)

        # Apply cutmix formula
        inputs_cutmixed = (inputs * mask_cutmix) + (random_partners * (1 - mask_cutmix))
        return inputs_cutmixed

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update({
            "probability": self.probability,
            "mask_seed": self.mask_seed,
            "shuffle_seed": self.shuffle_seed,
        })
        return config



from teras._src.layers.saint.multi_head_inter_sample_attention import SAINTMultiHeadInterSampleAttention
from keras.src.testing.test_case import TestCase
from keras import ops, random


class SAINTMultiHeadInterSampleAttentionTest(TestCase):
    def setUp(self):
        self.input_shape = (8, 5, 16)
        self.input_batch = random.normal((8, 5, 16))
        self.embedding_dim = self.input_shape[-1]

    def test_valid_output_shape(self):
        inter_sample_attention = SAINTMultiHeadInterSampleAttention(
            num_heads=4,
            key_dim=self.embedding_dim//8
        )
        outputs = inter_sample_attention(self.input_batch)
        self.assertEqual(ops.shape(outputs), self.input_shape)



from teras._src.backend.common.models.gans.ctgan.generator import BaseCTGANGenerator
from teras._src.typing import IntegerSequence


class CTGANGenerator(BaseCTGANGenerator):
    def __init__(self,
                 data_dim: int,
                 metadata: dict,
                 hidden_dims: IntegerSequence = (256, 256),
                 seed: int = 1337,
                 **kwargs):
        super().__init__(data_dim=data_dim,
                         metadata=metadata,
                         hidden_dims=hidden_dims,
                         seed=seed,
                         **kwargs)



import jax
import keras
from teras._src.backend.common.models.pretrainers.saint import BaseSAINTPretrainer


class SAINTPretrainer(BaseSAINTPretrainer):
    def __init__(self,
                 model: keras.Model,
                 cardinalities: list,
                 embedding_dim: int,
                 cutmix_probability: float = 0.3,
                 mixup_alpha: float = 1.,
                 temperature: float = 0.7,
                 lambda_: float = 10.,
                 lambda_c: float = 0.5,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         cardinalities=cardinalities,
                         embedding_dim=embedding_dim,
                         cutmix_probability=cutmix_probability,
                         mixup_alpha=mixup_alpha,
                         temperature=temperature,
                         lambda_=lambda_,
                         lambda_c=lambda_c,
                         seed=seed,
                         **kwargs)

    def compute_loss_and_updates(self,
                                 trainable_variables,
                                 non_trainable_variables,
                                 x,
                                 cardinalities,
                                 temperature,
                                 lambda_,
                                 lambda_c,
                                 training=False):
        (
            ((z_real, z_mixed), reconstructed),
            non_trainable_variables
        ) = self.stateless_call(trainable_variables,
                                non_trainable_variables,
                                x,
                                training=training)
        loss, c_loss, d_loss = self.compute_loss(
            x=x, x_reconstructed=reconstructed, z=z_real, z_augmented=z_mixed,
            cardinalities=cardinalities, temperature=temperature,
            lambda_c=lambda_c, lambda_=lambda_,
        )
        return loss, (c_loss, d_loss, reconstructed,
                      non_trainable_variables)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables
        ) = state

        grad_fn = jax.value_and_grad(self.compute_loss_and_updates,
                                     has_aux=True)

        (
            (loss, (c_loss, d_loss, reconstructed,
                    non_trainable_variables)),
            gradients
        ) = grad_fn(trainable_variables,
                    non_trainable_variables,
                    data,
                    self.cardinalities,
                    self.temperature,
                    self.lambda_,
                    self.lambda_c,
                    training=True)
        (
            trainable_variables,
            optimizer_variables
        ) = self.optimizer.stateless_apply(
            optimizer_variables,
            gradients,
            trainable_variables
        )

        # Update metrics
        logs = {}
        new_metric_variables = []
        for metric in self.metrics:
            this_metric_variables = metrics_variables[
                                    len(new_metric_variables): len(new_metric_variables) + len(metric.variables)
                                    ]
            if metric.name == "constrastive_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    c_loss
                )
            elif metric.name == "denoising_loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    d_loss
                )
            elif metric.name == "loss":
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    loss
                )
            else:
                this_metric_variables = metric.stateless_update_state(
                    this_metric_variables,
                    data,
                    reconstructed
                )
            logs[metric.name] = metric.stateless_result(this_metric_variables)
            new_metric_variables += this_metric_variables

        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            new_metric_variables
        )
        return logs, state



import tensorflow as tf
import pandas as pd
from typing import Union


def dataframe_to_tf_dataset(
        dataframe: pd.DataFrame,
        target: Union[str, list] = None,
        shuffle: bool = True,
        batch_size: int = 1024,
):
    """
    Builds a tf.data.Dataset from a given pandas dataframe

    Args:
        dataframe: `pd.DataFrame`,
            A pandas dataframe
        target: `str` or `list`,
            Name of the target column or list of names of the target columns.
        shuffle: `bool`, default True
            Whether to shuffle the dataset
        batch_size: `int`, default 1024,
            Batch size

    Returns:
         A tf.data.Dataset dataset
    """
    df = dataframe.copy()
    if target is not None:
        if isinstance(target, (list, tuple, set)):
            labels = []
            for feat in target:
                labels.append(df.pop(feat).values)
            labels = tf.transpose(tf.constant(labels))
        else:
            labels = df.pop(target)
            labels = labels.values
        dataset = tf.data.Dataset.from_tensor_slices((df.values, labels))
    else:
        dataset = tf.data.Dataset.from_tensor_slices(df.values)
    if shuffle:
        dataset = dataset.shuffle(buffer_size=len(dataframe))
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(batch_size)
    return dataset


def create_gain_dataset(x, seed: int = 1337):
    return tf.data.Dataset.from_tensor_slices(
        (x,
         tf.random.shuffle(x, seed=seed))
         )



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.CTGANGeneratorLayer")
class CTGANGeneratorLayer(keras.layers.Layer):
    """
    Residual Block for Generator as used by the authors of CTGAN
    proposed in the paper Modeling Tabular data using Conditional GAN.

    `outputs = Concat([ReLU(BatchNorm(Dense(inputs))), inputs])`

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        dim: int, Dimensionality of the hidden layer.
            Defaults to 256.
    """
    def __init__(self,
                 dim: int = 256,
                 **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self.dense = keras.layers.Dense(self.dim)
        self.batch_norm = keras.layers.BatchNormalization()
        self.relu = keras.layers.ReLU()

    def build(self, input_shape):
        self.dense.build(input_shape)
        input_shape = self.dense.compute_output_shape(input_shape)
        self.batch_norm.build(input_shape)
        self.relu.build(input_shape)

    def call(self, inputs):
        x = self.dense(inputs)
        x = self.batch_norm(x)
        x = self.relu(x)
        out = ops.concatenate([x, inputs], axis=1)
        return out

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (input_shape[-1] + self.dim,)

    def get_config(self):
        config = super().get_config()
        config.update({"dim": self.dim})
        return config



import keras
from teras._src.api_export import teras_export


@teras_export("teras.layers.CTGANDiscriminatorLayer")
class CTGANDiscriminatorLayer(keras.layers.Layer):
    """
    Discriminator Layer based on the architecture proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    outputs = Dropout(LeakyReLU(Dense(inputs)))

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        dim: int, Dimensionality of the hidden layer. Default to 256.
        leaky_relu_alpha: float, Alpha value to use for leaky relu activation.
            Defaults to 0.2
        dropout_rate: float, Dropout rate to use in the `Dropout` layer,
            which is applied after hidden layer. Defaults to 0.
    """
    def __init__(self,
                 dim: int = 256,
                 leaky_relu_alpha: float = 0.2,
                 dropout_rate: float = 0.,
                 **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self.leaky_relu_alpha = leaky_relu_alpha
        self.dropout_rate = dropout_rate

        self.dense = keras.layers.Dense(dim)
        self.leaky_relu = keras.layers.LeakyReLU(
            negative_slope=self.leaky_relu_alpha)
        self.dropout = keras.layers.Dropout(rate=self.dropout_rate)

    def build(self, input_shape):
        self.dense.build(input_shape)
        input_shape = self.dense.compute_output_shape(input_shape)
        self.leaky_relu.build(input_shape)
        self.dropout.build(input_shape)

    def call(self, inputs):
        out = self.dense(inputs)
        out = self.leaky_relu(out)
        out = self.dropout(out)
        return out

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            'dim': self.dim,
            'leaky_relu_alpha': self.leaky_relu_alpha,
            'dropout_rate': self.dropout_rate}
        )
        return config




import time
from abc import abstractmethod

import keras
from keras.backend import backend
from teras._src.backend.generic_utils import dataset_type

from teras._src.api_export import teras_export

_BACKEND = backend()


@teras_export("teras.models.GAN")
class GAN:
    """
    A base class for building specialized GAN-based architectures in Teras.

    Args:
        generator: An instance of `keras.Model` that would serve as the
            generator for the given GAN architecture.
        discriminator: An instance of `keras.Model` that would serve as
            the discriminator for the given GAN architecture.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model
                 ):
        self.generator = generator
        self.discriminator = discriminator

        self._trained = False

    @property
    def trained_generator(self):
        if not self._trained:
            raise AssertionError(
                "The generator has not yet been trained. Please train the "
                "GAN network using the `fit` method before accessing the "
                "`trained_generator` attribute."
            )
        return self.generator

    @property
    def trained_discriminator(self):
        if not self._trained:
            raise AssertionError(
                "The discriminator has not yet been trained. Please train the "
                "GAN network using the `fit` method before accessing the "
                "`trained_discriminator` attribute."
            )
        return self.discriminator

    @abstractmethod
    def generator_train_step(self, data):
        raise NotImplementedError

    @abstractmethod
    def discriminator_train_step(self, data):
        raise NotImplementedError

    def fit(self, x, batch_size: int = 128, epochs: int = 1,
            verbose: bool = True, shuffle: bool = True):
        """
        Fit method analogous to Keras's `Model.fit` method

        Args:
            x: Training dataset.
            batch_size: Batch size. Pass `None` to treat the whole dataset
                as one single batch.
            epochs: Number of epochs to train for.
            verbose: Whether to print outputs.
            shuffle: Whether to shuffle data. Only applicable when numpy
                ndarray is passed as the `x` argument.

        Returns:
            A dictionary containing the history of training, analogous
            to the history object returned by Keras's `fit` method.
        """

        # TODO:
        #   1. JAX ecosystem uses tf.data. Convert tf tensors to numpy
        #      arrays before passing to the model when on JAX backend.
        #   2. Create batches of data manually when passed a numpy ndarray
        #   3. Return a history object like Keras's fit method does

        total_batches = 0
        if dataset_type(x) == "not_supported":
            raise ValueError(
                "Unsupported type for `x`. "
                "It should be tensorflow dataset, pytorch dataloader or a "
                f"numpy ndarray. But received {type(x)}")
        elif dataset_type(x) == "numpy_ndarray":
            # TODO
            pass

        for epoch in range(epochs):
            epoch_start_time = time.time()
            for batch_num, batch in enumerate(x):
                generator_loss = self.discriminator_train_step(batch)
                discriminator_loss = self.generator_train_step(batch)

                if verbose:
                    epoch_str = f"Epoch {epoch + 1}/{epochs}"
                    elapsed_time_str = f"Elapsed {time.time() - epoch_start_time:.2f}s"
                    if total_batches > 0 and epoch > 0:
                        batch_str = f"Batch {batch_num + 1}/{total_batches}"
                    else:
                        batch_str = f"Batch {batch_num + 1}/?"
                        total_batches += 1
                    logs_str = f"generator_loss: {generator_loss:.4f}   " \
                               f"discriminator_loss: " \
                               f"{discriminator_loss:.4f}   "

                    print(
                        f"\r{epoch_str:<15} {elapsed_time_str:<15}"
                        f" {batch_str:<15}"
                        f" {logs_str}",
                        end="")
                print()



import tensorflow as tf
import keras
from keras import ops
from teras._src.backend.common.models.autoencoders.tvae.tvae import BaseTVAE


class TVAE(BaseTVAE):
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 metadata: dict,
                 data_dim: int,
                 latent_dim: int = 128,
                 loss_factor: float = 2.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(encoder=encoder,
                         decoder=decoder,
                         metadata=metadata,
                         data_dim=data_dim,
                         latent_dim=latent_dim,
                         loss_factor=loss_factor,
                         seed=seed,
                         **kwargs)

    def train_step(self, data):
        with tf.GradientTape() as tape:
            generated_samples, sigmas, mean, log_var = self(data)
            loss = self.compute_loss(real_samples=data,
                                     generated_samples=generated_samples,
                                     sigmas=sigmas,
                                     mean=mean,
                                     log_var=log_var)
        grads = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply(grads,
                             self.trainable_variables)
        self.decoder.sigmas.assign(ops.clip(sigmas, 0.01, 1.0))
        self.loss_tracker.update_state(loss)
        logs = {m.name: m.result() for m in self.metrics}
        return logs



import keras
from keras import random, ops
from teras._src.layers.ctgan.generator_layer import CTGANGeneratorLayer
from teras._src.layers.activation import GumbelSoftmax
from teras._src.typing import IntegerSequence
from teras._src.layers.layer_list import LayerList
from teras._src.utils import clean_reloaded_config_data


class BaseCTGANGenerator(keras.Model):
    """
    Base CTGANGenerator class.
    """
    def __init__(self,
                 data_dim: int,
                 metadata: dict,
                 hidden_dims: IntegerSequence = (256, 256),
                 seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_dim = data_dim
        self.metadata = metadata
        self.hidden_dims = hidden_dims

        self.hidden_block = []
        for dim in self.hidden_dims:
            self.hidden_block.append(CTGANGeneratorLayer(dim))
        self.hidden_block = LayerList(
            self.hidden_block,
            sequential=True,
            name="generator_hidden_block"
        )
        self.output_layer = keras.layers.Dense(
            self.data_dim,
            name="generator_output_layer")
        self.seed = seed
        self._seed_gen = random.SeedGenerator(self.seed)
        self.gumbel_softmax = GumbelSoftmax(seed=seed)

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self._input_shape = input_shape
        self.hidden_block.build(input_shape)
        input_shape = self.hidden_block.compute_output_shape(input_shape)
        self.output_layer.build(input_shape)
        self.gumbel_softmax.build(input_shape)

    def apply_activations_by_feature_type(self, interim_outputs):
        """
        This function applies activation functions to the interim outputs
        of the Generator by feature type.
        As CTGAN architecture requires specific transformations on the raw
        input data,
        that decompose one feature in several features,
        and since each type of feature, i.e. continuous or categorical
        require different activation functions to be applied, the process
        of applying those activations becomes rather tricky as it
        requires knowledge of underlying data transformation and
        features metadata.
        To ease the user's burden, in case a user wants to subclass this
        Generator model and completely customize the inner workings of the
        generator but would want to use the activation method specific
        to the CTGAN architecture, so that the subclassed Generator can
        work seamlessly with the rest of the architecture and there
        won't be any discrepancies in outputs produced by the subclasses
        Generator and those expected by the architecture,
        this function is separated, so user can just call this function on
        the interim outputs in the `call` method.

        Args:
            interim_outputs: Outputs produced by the `output_layer` of the
            Generator.

        Returns:
            Final outputs activated by the relevant activation functions.
        """
        outputs = []
        continuous_features_relative_indices = (
            self.metadata)["continuous"]["relative_indices_all"]
        features_relative_indices_all = (
            self.metadata)["relative_indices_all"]
        num_valid_clusters_all = (
            self.metadata)["continuous"]["num_valid_clusters_all"]
        cont_i = 0
        cat_i = 0
        num_categories_all = (
            self.metadata)["categorical"]["num_categories_all"]
        for i, index in enumerate(features_relative_indices_all):
            # the first k = num_continuous_features are continuous in the
            # data
            if i < len(continuous_features_relative_indices):
                # each continuous features has been transformed into
                # num_valid_clusters + 1 features
                # where the first feature is alpha while the following
                # features are beta components
                alphas = ops.tanh(interim_outputs[:, index])
                alphas = ops.expand_dims(alphas, 1)
                outputs.append(alphas)
                betas = self.gumbel_softmax(
                    interim_outputs[:, index + 1: index + 1 + num_valid_clusters_all[cont_i]])
                outputs.append(betas)
                cont_i += 1
            # elif index in categorical_features_relative_indices:
            else:
                # each categorical feature has been converted into a
                # one hot vector
                # of size num_categories
                ds = self.gumbel_softmax(
                    interim_outputs[:, index: index + num_categories_all[cat_i]])
                outputs.append(ds)
                cat_i += 1
        outputs = ops.concatenate(outputs, axis=1)
        return outputs

    def call(self, inputs):
        # inputs have the shape |z| + |cond|
        # while the outputs will have the shape of equal to
        # (batch_size, transformed_data_dims)
        interim_outputs = self.output_layer(self.hidden_block(inputs))
        outputs = self.apply_activations_by_feature_type(interim_outputs)
        return outputs

    def predict_step(self, inputs):
        generated_samples = self(inputs)
        return generated_samples

    def compute_output_shape(self, input_shape):
        input_shape = tuple(input_shape)
        return input_shape[:-1] + (self.data_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            'data_dim': self.data_dim,
            'metadata': self.metadata,
            'hidden_dims': self.hidden_dims,
        }
        )
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        metadata = clean_reloaded_config_data(config.pop("metadata"))
        return cls(data_dim=data_dim,
                   metadata=metadata,
                   **config)



import keras

from teras._src.layers.layer_list import LayerList
from teras._src.typing import IntegerSequence
from teras._src.api_export import teras_export


@teras_export("teras.models.TVAEEncoder")
class TVAEEncoder(keras.Model):
    """
    Encoder for the TVAE model as proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        latent_dim: int, Dimensionality of the learned latent space.
            Defaults to 128.
        compression_dims: Sequence, A sequence of integers. For each value in 
            the sequence, a dense layer of that dimensions is added to 
            construct a compression block.
            Defaults to (128, 128).
    """
    def __init__(self,
                 latent_dim: int = 128,
                 compression_dims: IntegerSequence = (128, 128),
                 **kwargs):
        super().__init__(**kwargs)

        if not isinstance(compression_dims, (list, tuple)):
            raise ValueError(
                f"`compression_dims` must be a sequence of integers."
                f"Received: {compression_dims}")

        self.latent_dim = latent_dim
        self.compression_dims = compression_dims

        self.compression_block = []
        for i, units in enumerate(self.compression_dims, start=1):
            self.compression_block.append(
                keras.layers.Dense(units=units,
                                   activation="relu",
                                   name=f"compression_layer_{i}"))
        self.compression_block = LayerList(
            self.compression_block,
            sequential=True,
            name="tvae_encoder_compression_block"
        )
        self.dense_mean = keras.layers.Dense(self.latent_dim,
                                             name="mean")
        self.dense_log_var = keras.layers.Dense(self.latent_dim,
                                                name="log_var")

    def build(self, input_shape):
        self.compression_block.build(input_shape)
        input_shape = self.compression_block.compute_output_shape(input_shape)
        self.dense_mean.build(input_shape)
        self.dense_log_var.build(input_shape)

    def call(self, inputs):
        h = self.compression_block(inputs)
        mean = self.dense_mean(h)
        log_var = self.dense_log_var(h)
        return mean, log_var

    def compute_output_shape(self, input_shape):
        batch_size, dims = input_shape
        return ((batch_size, self.latent_dim),
                (batch_size, self.latent_dim))

    def get_config(self):
        config = super().get_config()
        config.update({
            'latent_dim': self.latent_dim,
            'compression_dims': self.compression_dims
        })
        return config




import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.SAINTMultiHeadInterSampleAttention")
class SAINTMultiHeadInterSampleAttention(keras.layers.Layer):
    """
    Multi Head Inter Sample Attention layer based on the SAINT architecture
    proposed in the "SAINT: Improved Neural Networks for Tabular Data"
    paper.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        num_heads: int, number of attention heads to use.
        key_dim: int, the paper proposes to use embedding_dim/num_heads
            dimensions for your key dimensionality
        value_dim: int, same value as key_dim is used by the paper.
        dropout: float, dropout value to use. Defaults to 0.

    Shapes:
        Input Shape: (batch_size, num_features, embedding_dim)
        Output Shape: (batch_size, num_features, embedding_dim)
    """
    def __init__(self,
                 num_heads: int,
                 key_dim: int,
                 value_dim: int = None,
                 dropout: float = 0.0,
                 **kwargs):
        super().__init__(**kwargs)
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.value_dim = value_dim
        self.dropout = dropout

        self.multi_head_attention = keras.layers.MultiHeadAttention(
            num_heads=self.num_heads,
            key_dim=self.key_dim,
            value_dim=self.value_dim,
            dropout=self.dropout,
        )

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise ValueError(
                "Inputs must have the shape of rank 3"
                "(`batch_size`, `num_features`, embedding_dim`) but "
                f"received shape {input_shape} with rank "
                f"{len(input_shape)}."
            )
        input_shape = (1, input_shape[0], input_shape[1] * input_shape[2])
        self.multi_head_attention.build(input_shape, input_shape)

    def call(self, inputs):
        batch_size, num_features, embedding_dim = ops.shape(inputs)
        x = ops.reshape(inputs,
                        (1, batch_size, num_features * embedding_dim),
                        )
        x = self.multi_head_attention(x, x)
        x = ops.reshape(x,
                        (batch_size, num_features, embedding_dim))
        return x

    def compute_output_shape(self, input_shape):
        return input_shape  # easy!

    def get_config(self):
        config = super().get_config()
        config.update({
            "num_heads": self.num_heads,
            "key_dim": self.key_dim,
            "value_dim": self.value_dim,
            "dropout": self.dropout
        })



import pandas as pd
import numpy as np
from sklearn.preprocessing import OrdinalEncoder
from teras._src.preprocessing.data_transformers.data_transformer import DataTransformer
from teras._src.typing import FeaturesNamesType
import json


class GAINDataTransformer(DataTransformer):
    """
    GAINDataTransformer class that performs the required transformations
    on the raw dataset required by the GAIN architecture.

    Args:
        categorical_features: list, List of categorical features names in the
            dataset. Categorical features are encoded by ordinal encoder method.
            And then MinMax normalization is applied.
        continuous_features: list, List of numerical features names
            in the dataset. Numerical features are encoded using MinMax
            normalization.
    """
    def __init__(self,
                 categorical_features: FeaturesNamesType = None,
                 continuous_features: FeaturesNamesType = None
                 ):
        super().__init__()
        self.categorical_features = categorical_features
        self.continuous_features = continuous_features
        self._encoder = OrdinalEncoder()
        self._min_vals = None
        self._max_vals = None
        self._ordered_features_names_all = None
        self._fitted = False

    def fit(self, x: pd.DataFrame):
        # min-max normalize the entire dataset regardless of the features types.
        # For categorical, first need to encode values ordinally
        x_temp = x.copy()
        if self.categorical_features is not None:
            self._encoder.fit(x[self.categorical_features])
            x_temp[self.categorical_features] = self._encoder.transform(
                x[self.categorical_features])
        self._min_vals = np.nanmin(x_temp, axis=0)
        self._max_vals = np.nanmax(x_temp, axis=0)
        self._fitted = True

    def transform(self, x: pd.DataFrame):
        """
        Transforms the data (applying normalization etc)
        and returns a tensorflow dataset.
        It also stores the metadata of features
        that is used in the reverse transformation step.

        Args:
            x: Data to transform. Must be a pandas DataFrame.

        Returns:
            Transformed data.
        """
        if not self._fitted:
            raise AssertionError(
                "You haven't yet fitted the DataTransformer. "
                "You must call the `fit` method before you can call the "
                "`transform` method. ")
        if not isinstance(x, pd.DataFrame):
            raise ValueError(
                "Only pandas dataframe is supported by DataTransformer class."
                f" But data of type {type(x)} was passed. "
                f"Please convert it to pandas dataframe before passing.")
        self._ordered_features_names_all = x.columns

        if self.categorical_features is not None:
            x[self.categorical_features] = self._encoder.transform(
                x[self.categorical_features])

        x = (x - self._min_vals) / self._max_vals
        return x

    def inverse_transform(self, x):
        """
        Inverse Transforms the transformed data.

        Args:
            x: Transformed Data.

        Returns:
            Pandas dataframe of data in its original scale
        """
        if not isinstance(x, pd.DataFrame):
            x = pd.DataFrame(x, columns=self._ordered_features_names_all)

        if not self._fitted:
            raise AssertionError(
                "You haven't yet fitted the DataTransformer. "
                "You must call the `fit` method before you can call the "
                "`inverse_transform` method. ")

        # min-max transformation was applied to the whole dataset even to the
        # ordinal encoded categorical features, so first, undo that
        # transformation and then inverse transform categorical
        x = (x * self._max_vals) + self._min_vals
        if self.categorical_features is not None:
            x[self.categorical_features] = self._encoder.inverse_transform(
                x[self.categorical_features])
        return x

    def save(self, filename):
        """
        Saves the fitted state of `DataTransformer` instance for portability,
        in the `json` format.

        Args:
            filename: Filename or file path ending in `.json` extension.
        """
        args = {
            "categorical_features": self.categorical_features,
            "continuous_features": self.continuous_features
        }
        attrs = {
            "_min_vals": list(self._min_vals),
            "_max_vals": list(self._max_vals),
            "_ordered_features_names_all": self._ordered_features_names_all,
            "_fitted": self._fitted
        }
        encoder_attrs = {
            "categories_": list(self._encoder.categories_)
        }
        state = {
            "args": args,
            "attrs": attrs,
            "encoder_attrs": encoder_attrs,
        }
        with open(filename, "w") as f:
            json.dump(state, f)

    @classmethod
    def load(cls, filename):
        """
        Loads the saved state of `DataTransformer` from the `json` file.

        Args:
            filename: Filename or file path ending in `.json` extension.

        Returns:
            An instance of `GAINDataTransformer` with state stored in the
            `filename` json file.
        """
        with open(filename, "r") as f:
            state = json.load(f)
        c = cls(**state.pop("params"))
        for name, value in state.pop("attrs"):
            c.__setattr__(name, value)
        for name, value in state.pop("encoder_attrs"):
            c._encoder.__setattr__(name, value)
        return c



from teras._src.layers.saint.projection_head import SAINTProjectionHead
from keras.src.testing.test_case import TestCase
from keras import ops, random


class SAINTProjectionHeadTest(TestCase):
    def setUp(self):
        self.input_batch = random.normal((8, 5, 16))
        self.embedding_dim = 16

    def test_valid_call(self):
        projection_head = SAINTProjectionHead(
            hidden_dim=16,
            output_dim=10,
        )
        outputs = projection_head(self.input_batch)

    def test_valid_output_shape(self):
        projection_head = SAINTProjectionHead(
            hidden_dim=16,
            output_dim=10,
        )
        outputs = projection_head(self.input_batch)
        self.assertEqual(ops.shape(outputs), (8, 5, 10))



from teras._src.layers.tab_transformer.column_embedding import TabTransformerColumnEmbedding
from keras import ops, random
from keras.src.testing.test_case import TestCase


class TabTransformerColumnEmbeddingTest(TestCase):
    def setUp(self):
        # we assume there are 4 features in the dataset, with feature 1
        # having 5 unique categories, feature 2 having 3 unique
        # categories, feature 3 with 2 unique categories and feature 4
        # with 7 unique categories
        x_cat = ops.array([
            ops.cast(random.randint((16,), minval=0, maxval=card),
                     dtype="float32")
            for card in [5, 3, 2, 7]]
        )
        x_cat = ops.transpose(x_cat)
        x_cont = random.normal((16, 3), dtype="float32")
        self.input_batch = ops.concatenate([x_cat, x_cont], axis=1)
        self.cardinalities = [5, 3, 2, 7, 0, 0, 0]  # 0 for continuous

    def test_valid_output_shape(self):
        column_embedding = TabTransformerColumnEmbedding(
            cardinalities=self.cardinalities,
            embedding_dim=32,
        )
        embeddings = column_embedding(self.input_batch)
        self.assertEqual((16, 4, 32), ops.shape(embeddings))

    def test_with_custom_shared_embedding_dim(self):
        column_embedding = TabTransformerColumnEmbedding(
            cardinalities=self.cardinalities,
            embedding_dim=32,
            shared_embedding_dim=8,
        )
        embeddings = column_embedding(self.input_batch)
        self.assertEqual((16, 4, 32), ops.shape(embeddings))

    def test_when_use_shared_embedding_is_false(self):
        column_embedding = TabTransformerColumnEmbedding(
            cardinalities=self.cardinalities,
            embedding_dim=32,
            use_shared_embedding=False,
        )
        embeddings = column_embedding(self.input_batch)
        self.assertEqual((16, 4, 32), ops.shape(embeddings))

    def test_when_join_method_is_add(self):
        column_embedding = TabTransformerColumnEmbedding(
            cardinalities=self.cardinalities,
            embedding_dim=32,
            join_method="add"
        )
        embeddings = column_embedding(self.input_batch)
        self.assertEqual((16, 4, 32), ops.shape(embeddings))



import tensorflow as tf
import keras
from keras import ops, random
from teras._src.backend.common.models.pretrainers.tabnet import BaseTabNetPretrainer


class TabNetPretrainer(BaseTabNetPretrainer):
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 missing_feature_probability: float = 0.3,
                 **kwargs):
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            missing_feature_probability=missing_feature_probability,
            **kwargs)

    def train_step(self, data):
        if isinstance(data, tuple):
            data = data[0]
        # Sample mask
        mask = random.binomial(
            shape=ops.shape(data),
            counts=1,
            probabilities=self.missing_feature_probability,
            seed=self._seed_generator
        )
        with tf.GradientTape() as tape:
            reconstructed = self(data, mask, training=True)
            loss = self.compute_loss(x=data,
                                     x_reconstructed=reconstructed,
                                     mask=mask)
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply(
            grads=gradients,
            trainable_variables=self.trainable_variables
        )

        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(data, reconstructed)

        logs = {m.name: m.result() for m in self.metrics}
        return logs



import os

import keras
from keras import random, ops
from keras.src.testing.test_case import TestCase
from teras._src.models.pretrainers.tabnet.decoder import TabNetDecoder


class TabNetDecoderTest(TestCase):
    def setUp(self):
        self.batch_size = 16
        self.encoded_features = random.normal((self.batch_size, 8))

    def test_valid_call(self):
        model = TabNetDecoder(data_dim=5,
                              feature_transformer_dim=16,
                              decision_step_dim=8)
        reconstructed_features = model(self.encoded_features)

    def test_valid_output_shape(self):
        model = TabNetDecoder(data_dim=5,
                              feature_transformer_dim=16,
                              decision_step_dim=8)
        reconstructed_features = model(self.encoded_features)
        self.assertEqual(ops.shape(reconstructed_features),
                         (self.batch_size, 5))

    def test_model_save_and_load(self):
        model = TabNetDecoder(data_dim=5,
                              feature_transformer_dim=16,
                              decision_step_dim=8)
        reconstructed_features = model(self.encoded_features)
        save_path = os.path.join(self.get_temp_dir(),
                                 "tabnet_decoder_backbone.keras"
                                 )
        model.save(save_path)
        reloaded_model = keras.models.load_model(save_path)

        # Check we got the real object back
        self.assertIsInstance(reloaded_model, TabNetDecoder)

        # Check that output matches
        reloaded_outputs = reloaded_model(self.encoded_features)
        self.assertAllClose(
            ops.convert_to_numpy(reconstructed_features),
            ops.convert_to_numpy(reloaded_outputs)
        )



from keras import ops, random
from teras._src.layers.categorical_embedding import CategoricalEmbedding
from keras.src.testing.test_case import TestCase


class CategoricalEmbeddingTest(TestCase):
    def setUp(self):
        self.input_batch = ops.array([ops.repeat(ops.array([1., 2.]), 5),
                                      ops.arange(0., 10.),
                                      random.normal((10,)),
                                      ])
        self.input_batch = ops.transpose(self.input_batch)
        # 0 represents the numerical features as required by the
        # `CategoricalEmbedding` layer
        self.cardinalities = [2, 10, 0]

    def test_valid_output_shape(self):
        ce = CategoricalEmbedding(embedding_dim=16,
                                  cardinalities=self.cardinalities)
        embeddings = ce(self.input_batch)
        expected_shape = (10, 2, 16) # (batch_size, num_categorical_features, embedding_dim)
        self.assertEqual(expected_shape, ops.shape(embeddings))

    def test_valid_output_shape_when_only_one_categorical(self):
        input_batch = ops.array([random.normal((10,)),
                                 random.normal((10,)),
                                 ops.repeat(ops.array([1., 2.]), 5),
                                 ])
        input_batch = ops.transpose(input_batch)
        cardinalities = [0, 0, 2]
        ce = CategoricalEmbedding(embedding_dim=16,
                                  cardinalities=cardinalities)
        embeddings = ce(input_batch)
        expected_shape = (10, 1, 16)
        self.assertEqual(expected_shape, ops.shape(embeddings))



# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Backbones
from teras._src.models.backbones.transformer.encoder import (
    TransformerEncoderBackbone as TransformerEncoderBackbone)
from teras._src.models.backbones.ft_transformer.ft_transformer import (
    FTTransformerBackbone as FTTransformerBackbone)
from teras._src.models.backbones.tab_transformer.tab_transformer import (
    TabTransformerBackbone as TabTransformerBackbone)
from teras._src.models.backbones.saint.saint import (
    SAINTBackbone as SAINTBackbone)
from teras._src.models.backbones.tabnet.encoder import (
    TabNetEncoderBackbone as TabNetEncoderBackbone)


# Pretrainers
from teras._src.models.pretrainers.tab_transformer import (
    TabTransformerRTDPretrainer as TabTransformerRTDPretrainer,
    TabTransformerMLMPretrainer as TabTransformerMLMPretrainer
)
from teras._src.models.pretrainers.saint import (
    SAINTPretrainer as SAINTPretrainer)
from teras._src.models.pretrainers.tabnet.tabnet import (
    TabNetPretrainer as TabNetPretrainer)
from teras._src.models.pretrainers.tabnet.decoder import (
    TabNetDecoder as TabNetDecoder
)


# GANs
from teras._src.models.gans.gain.gain import GAIN as GAIN
from teras._src.models.gans.gain.generator import GAINGenerator as GAINGenerator
from teras._src.models.gans.gain.discriminator import (
    GAINDiscriminator as GAINDiscriminator)
from teras._src.models.gans.pcgain.pcgain import PCGAIN as PCGAIN
from teras._src.models.gans.ctgan.ctgan import CTGAN as CTGAN
from teras._src.models.gans.ctgan.generator import (
    CTGANGenerator as CTGANGenerator)
from teras._src.models.gans.ctgan.discriminator import (
    CTGANDiscriminator as CTGANDiscriminator)


# Autoencoders
from teras._src.models.autoencoders.tvae.tvae import TVAE as TVAE
from teras._src.models.autoencoders.tvae.encoder import (
    TVAEEncoder as TVAEEncoder)
from teras._src.models.autoencoders.tvae.decoder import (
    TVAEDecoder as TVAEDecoder)


# Task models
from teras._src.models.tasks.regression import Regressor as Regressor
from teras._src.models.tasks.classification import Classifier as Classifier



import tensorflow as tf
import keras
from teras._src.backend.common.models.pretrainers.saint import BaseSAINTPretrainer


class SAINTPretrainer(BaseSAINTPretrainer):
    def __init__(self,
                 model: keras.Model,
                 cardinalities: list,
                 embedding_dim: int,
                 cutmix_probability: float = 0.3,
                 mixup_alpha: float = 1.,
                 temperature: float = 0.7,
                 lambda_: float = 10.,
                 lambda_c: float = 0.5,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         cardinalities=cardinalities,
                         embedding_dim=embedding_dim,
                         cutmix_probability=cutmix_probability,
                         mixup_alpha=mixup_alpha,
                         temperature=temperature,
                         lambda_=lambda_,
                         lambda_c=lambda_c,
                         seed=seed,
                         **kwargs)

    def train_step(self, data):
        with tf.GradientTape() as tape:
            (z_real, z_mixed), reconstructed = self(data)
            loss, c_loss, d_loss = self.compute_loss(
                x=data, x_reconstructed=reconstructed, z=z_real,
                z_augmented=z_mixed, cardinalities=self.cardinalities,
                temperature=self.temperature, lambda_=self.lambda_,
                lambda_c=self.lambda_c
            )
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply(gradients,
                             self.trainable_variables)

        for metric in self.metrics:
            if metric.name == "constrastive_loss":
                metric.update_state(c_loss)
            elif metric.name == "denoising_loss":
                metric.update_state(d_loss)
            elif metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(data, reconstructed)

        logs = {m.name: m.result() for m in self.metrics}
        return logs



import os
os.environ["KERAS_BACKEND"] = "jax"
import tensorflow as tf
from teras._src.models.gans.gain.generator import GAINGenerator
from teras._src.models.gans.gain.discriminator import GAINDiscriminator
import numpy as np
from teras._src.utils import inject_missing_values
from teras._src.backend.jax.trainers import gain as GAINTrainer


data_dim = 7
rng = np.random.default_rng(1337)
x_gen = rng.uniform(0., high=10., size=(128, data_dim)).astype(np.float32)
x_gen = inject_missing_values(x_gen)
x_disc = rng.uniform(0., high=10., size=(128, data_dim)).astype(np.float32)
x_disc = inject_missing_values(x_disc)
input_ds = tf.data.Dataset.from_tensor_slices((x_gen, x_disc))
input_ds = input_ds.batch(8)

generator = GAINGenerator(data_dim=data_dim)
discriminator = GAINDiscriminator(data_dim=data_dim)


GAINTrainer.init(generator, discriminator)
GAINTrainer.compile()

gen_state, disc_state = GAINTrainer.fit(input_ds, epochs=10)



from keras import ops
from teras._src.layers.mixup import MixUp
from keras.src.testing.test_case import TestCase


class MixUpTest(TestCase):
    def setUp(self):
        self.input_batch = ops.ones((8, 4))

    def test_mixup_valid_call(self):
        mixup = MixUp()
        outputs = mixup(self.input_batch)

    def test_mixup_output_shape(self):
        mixup = MixUp()
        outputs = mixup(self.input_batch)
        self.assertEqual(ops.shape(outputs),
                         ops.shape(self.input_batch))



import torch
from torch.utils.data import Dataset


class GAINDataset(Dataset):
    def __init__(self, x, seed: int = 1337):
        self.x = x
        self.num_samples = x.shape[0]
        self.seed = seed
        torch.manual_seed(self.seed)
        self.random_idx = torch.randperm(self.num_samples)

    def __getitem__(self, item):
        return self.x[item], self.x[self.random_idx][item]

    def __len__(self):
        return self.num_samples


def create_gain_dataset(x, seed: int = 1337):
    gain_ds = GAINDataset(x, seed=seed)
    return gain_ds



import os
os.environ["KERAS_BACKEND"] = "jax"
import keras
import tensorflow as tf
from teras._src.backend.jax.models.gans.pcgain import PCGAIN
from teras._src.models.gans.gain.generator import GAINGenerator
from teras._src.models.gans.gain.discriminator import GAINDiscriminator
from teras._src.models.gans.pcgain.classifier import PCGAINClassifier
import numpy as np
from teras._src.utils import inject_missing_values


DATA_DIM = 7

NUM_SAMPLES = 256
BATCH_SIZE = 32

rng = np.random.default_rng(1337)
x_gen = rng.uniform(0., high=10., size=(NUM_SAMPLES, DATA_DIM)).astype(np.float32)
x_gen = inject_missing_values(x_gen)
x_disc = rng.uniform(0., high=10., size=(NUM_SAMPLES, DATA_DIM)).astype(np.float32)
x_disc = inject_missing_values(x_disc)
input_ds = tf.data.Dataset.from_tensor_slices((x_gen, x_disc))
input_ds = input_ds.batch(BATCH_SIZE)

generator = GAINGenerator(data_dim=DATA_DIM)
discriminator = GAINDiscriminator(data_dim=DATA_DIM)
classifier = PCGAINClassifier(num_classes=4,
                              data_dim=DATA_DIM)
pcgain = PCGAIN(generator,
                discriminator,
                classifier)
pcgain.build(tuple(x_gen.shape))
pcgain.compile(generator_optimizer=keras.optimizers.Adam(),
               discriminator_optimizer=keras.optimizers.Adam())
pcgain.build_optimizers()
history = pcgain.fit(input_ds, epochs=10)



import keras

from teras._src.layers.cls_token import CLSToken
from teras._src.layers.cls_token_extraction import CLSTokenExtraction
from teras._src.layers.ft_transformer.feature_tokenizer import \
    FTTransformerFeatureTokenizer
from teras._src.models.backbones.backbone import Backbone
from teras._src.models.backbones.transformer.encoder import \
    TransformerEncoderBackbone
from teras._src.api_export import teras_export


@teras_export("teras.models.FTTransformerBackbone")
class FTTransformerBackbone(Backbone):
    """
    FT-Transformer Encoder backbone based on the FT-Transformer
    architecture proposed in the
    "Revisiting Deep Learning Models for Tabular Data" paper.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        input_dim: int, dimensionality of the input data.
        cardinalities: list, a list cardinalities of all the features
            in the dataset in the same order as the features' occurrence.
            For numerical features, use any value <=0 as indicator at
            the corresponding index.
            You can use the `compute_cardinalities` function from
            `teras.utils` package for this purpose.
        embedding_dim: int, dimensionality of the embeddings used
            by the model. It is also referred to as the `d_model` or
            model dimensionality.
        num_layers: int, number of `TransformerEncoderLayer`s to use in
            the encoder.
        num_heads: int, number of attention heads to use in the
            `MultiHeadAttention` layer.
        feedforward_dim: int, hidden dimensionality to use in the
            `TransformerFeedForward` layer.
        attention_dropout: float, dropout value to use in the
        `MultiHeadAttention` layer. Defaults to 0.
        feedforward_dropout: float, dropout value to use in the
            `TransformerFeedForward` layer. Defaults to 0.
        layer_norm_epsilon: float, epsilon value to use in the
            `LayerNormalization` layer. Defaults to 1e-5.
    """
    def __init__(self,
                 input_dim: int,
                 cardinalities: list,
                 embedding_dim: int,
                 num_layers: int = 6,
                 num_heads: int = 8,
                 feedforward_dim: int = None,
                 attention_dropout: float = 0.,
                 feedforward_dropout: float = 0.,
                 layer_norm_epsilon: float = 1e-5,
                 **kwargs):
        inputs = keras.layers.Input(shape=(input_dim,), name="inputs")
        x = FTTransformerFeatureTokenizer(
            cardinalities=cardinalities,
            embedding_dim=embedding_dim)(inputs)
        x = CLSToken(embedding_dim=embedding_dim)(x)
        x = TransformerEncoderBackbone(
            input_dim=input_dim + 1,    # Add plus one for the cls token
            embedding_dim=embedding_dim,
            num_layers=num_layers,
            num_heads=num_heads,
            feedforward_dim=feedforward_dim,
            attention_dropout=attention_dropout,
            feedforward_dropout=feedforward_dropout,
            layer_norm_epsilon=layer_norm_epsilon,
            unnormalized_layers=[0],
            pre_normalization=True)(x)
        outputs = CLSTokenExtraction()(x)
        super().__init__(inputs=inputs, outputs=outputs, **kwargs)

        self.input_dim = input_dim
        self.cardinalities = cardinalities
        self.num_layers = num_layers
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.feedforward_dim = feedforward_dim
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.layer_norm_epsilon = layer_norm_epsilon

    def get_config(self):
        config = super().get_config()
        config.update({
            "input_dim": self.input_dim,
            "cardinalities": self.cardinalities,
            "embedding_dim": self.embedding_dim,
            "num_layers": self.num_layers,
            "num_heads": self.num_heads,
            "feedforward_dim": self.feedforward_dim,
            "attention_dropout": self.attention_dropout,
            "feedforward_dropout": self.feedforward_dropout,
            "layer_norm_epsilon": self.layer_norm_epsilon,
        })
        return config



import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.TabNetPretrainer")
class TabNetPretrainer(backend.models.TabNetPretrainer):
    """
    TabNetPretrainer for pretraining `TabNetEncoder` as proposed by
    Arik et al. in the paper,
    "TabNet: Attentive Interpretable Tabular Learning"

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        encoder: keras.Model, instance of `TabNetEncoder` to pretrain
        decoder: keras.Model, instance of `TabNetDecoder`
        missing_feature_probability: float, probability of missing features
        seed: int, seed for generating mask. Defaults to 1337
    """
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 missing_feature_probability: float = 0.3,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            missing_feature_probability=missing_feature_probability,
            seed=seed,
            **kwargs)
